@@ -3,12 +3,12 @@ use base 'XML::Compile::Translate';
 
 use strict;
 use warnings;
-no warnings 'once';
+no warnings 'once', 'recursion';
 
 use Log::Report 'xml-compile', syntax => 'SHORT';
 use List::Util qw/first/;
 
-use XML::Compile::Util qw/pack_type odd_elements type_of_node/;
+use XML::Compile::Util qw/pack_type odd_elements type_of_node SCHEMA2001i/;
 use XML::Compile::Iterator ();
 
 =chapter NAME
@@ -41,14 +41,9 @@ into a (nested) Perl HASH structure.
 # that the structure of found data is not conforming the needs. For optional
 # blocks, these errors are caught and un-done.
 
-sub actsAs($) { $_[1] eq 'READER' }
-
-sub makeTagUnqualified
-{   my ($self, $path, $node, $local, $ns) = @_;
-    $local =~ s/.*?\://;   # strip prefix, that's all
-    $local;
-}
-*makeTagQualified = \&makeTagUnqualified;
+sub actsAs($)             {$_[1] eq 'READER'}
+sub makeTagUnqualified(@) {$_[3]} # ($self, $path, $node, $local, $ns)
+sub makeTagQualified(@)   {$_[3]} # same params
 
 sub typemapToHooks($$)
 {   my ($self, $hooks, $typemap) = @_;
@@ -83,7 +78,7 @@ sub typemapToHooks($$)
             $hook = sub {$object->fromXML($_[1], $type)};
         }
 
-        push @$hooks, { type => $type, after => $hook };
+        push @$hooks, +{action => 'READER', type => $type, after => $hook};
     }
     $hooks;
 }
@@ -104,7 +99,13 @@ sub makeElementWrapper
                   sub { $_[0]->isa('XML::LibXML::Element') } );
           }
 
-          ($processor->($tree))[-1];
+          my $data = ($processor->($tree))[-1];
+          unless(defined $data)
+          {    my $node = $tree->node;
+               error __x"data not recognized, found a `{type}' at {where}"
+                  , type => type_of_node $node, where => $node->nodePath;
+          }
+          $data;
         };
 }
 
@@ -124,7 +125,7 @@ sub makeAttributeWrapper
 }
 
 sub makeWrapperNs        # no namespaces in the HASH
-{   my ($self, $path, $processor, $index) = @_;
+{   my ($self, $path, $processor, $index, $filter) = @_;
     $processor;
 }
 
@@ -198,7 +199,7 @@ sub makeChoice($@)
           # a choice, instead on choice itself.  That always succeeds.
           foreach my $some (values %do)
           {   try { $some->(undef) };
-              $@ or return;
+              $@ or return ();
           }
 
           $type
@@ -322,10 +323,10 @@ sub makeBlockHandler
     # not be an additional nesting in the output tree.
     if($max ne 'unbounded' && $max==1)
     {
-       return ($label => $process) if $min==1;
+        return ($label => $process) if $min==1;
 
         my $code =
-        sub { my $tree = shift or return ();
+        sub { my $tree    = shift or return ();
               my $starter = $tree->currentChild or return ();
               my @pairs   = try { $process->($tree) };
               if($@->wasFatal(class => 'misfit'))
@@ -411,7 +412,7 @@ sub makeBlockHandler
 
 sub makeElementHandler
 {   my ($self, $path, $label, $min, $max, $required, $optional) = @_;
-    $max eq "0" and return sub {};
+    $max eq "0" and return sub {};  # max can be "unbounded"
 
     if($max ne 'unbounded' && $max==1)
     {   return $min==1
@@ -504,7 +505,6 @@ sub makeElementHref
 
 sub makeElement
 {   my ($self, $path, $ns, $childname, $do) = @_;
-
     sub { my $tree  = shift;
           my $value = defined $tree && $tree->nodeType eq $childname
              ? $do->($tree) : $do->(undef);
@@ -570,29 +570,8 @@ sub makeElementFixed
         };
 }
 
-sub makeElementNillable
-{   my ($self, $path, $ns, $childname, $do) = @_;
-
-    # some people cannot read the specs.
-    my $inas = $self->{interpret_nillable_as_optional};
-
-    sub { my $tree = shift;
-          my $value;
-          if(defined $tree && $tree->nodeType eq $childname)
-          {   my $nil  = $tree->node->getAttribute('nil') || 'false';
-              return ($childname => 'NIL')
-                  if $nil eq 'true' || $nil eq '1';
-              $value = $do->($tree);
-          }
-          elsif($inas) { return ($childname => undef) }
-          else { $value = $do->(undef) }
-
-          defined $value ? ($childname => $value) : ();
-        };
-}
-
 sub makeElementAbstract
-{   my ($self, $path, $ns, $childname, $do) = @_;
+{   my ($self, $path, $ns, $childname, $do, $tag) = @_;
     sub { my $tree = shift or return ();
           $tree->nodeType eq $childname or return ();
 
@@ -605,23 +584,48 @@ sub makeElementAbstract
 # complexType and complexType/ComplexContent
 #
 
+sub _notProcessed($$)
+{   my ($self, $child, $path) = @_;
+    error __x"element `{name}' not processed for {path} at {where}"
+      , name => type_of_node($child), path => $path
+      , _class => 'misfit', where => $child->nodePath;
+}
+
 sub makeComplexElement
-{   my ($self, $path, $tag, $elems, $attrs, $attrs_any) = @_;
-my @e = @$elems;
+{   my ($self, $path, $tag, $elems, $attrs, $attrs_any,undef,$is_nillable) = @_;
+#my @e = @$elems; my @a = @$attrs;
     my @elems = odd_elements @$elems;
     my @attrs = (odd_elements(@$attrs), @$attrs_any);
 
+    $is_nillable || @elems > 1 || @attrs and return
     sub { my $tree    = shift or return ();
           my $node    = $tree->node;
-          my %complex
-           = ( (map {$_->($tree)} @elems)
-             , (map {$_->($node)} @attrs)
-             );
+          my %complex =
+            ( ($tree->nodeNil ? (_ => 'NIL') : (map $_->($tree), @elems))
+            , (map $_->($node), @attrs)
+            );
 
-          defined $tree->currentChild
-              and error __x"element `{name}' not processed at {path}"
-                      , name => $tree->currentType, path => $path
-                      , _class => 'misfit';
+          $self->_notProcessed($tree->currentChild, $path)
+             if $tree->currentChild;
+
+          ($tag => \%complex);
+        };
+
+    @elems || return
+    sub { my $tree = shift or return ();
+
+          $self->_notProcessed($tree->currentChild, $path)
+             if $tree->currentChild;
+
+          ($tag => {});
+        };
+
+    my $el = shift @elems;
+    sub { my $tree    = shift or return ();
+          my %complex = $el->($tree);
+
+          $self->_notProcessed($tree->currentChild, $path)
+             if $tree->currentChild;
 
           ($tag => \%complex);
         };
@@ -632,16 +636,15 @@ my @e = @$elems;
 #
 
 sub makeTaggedElement
-{   my ($self, $path, $tag, $st, $attrs, $attrs_any) = @_;
+{   my ($self, $path, $tag, $st, $attrs, $attrs_any,undef,$is_nillable) = @_;
     my @attrs = (odd_elements(@$attrs), @$attrs_any);
 
     sub { my $tree   = shift or return ();
-          my $simple = $st->($tree);
+          my $simple = $is_nillable && $tree->nodeNil ? 'NIL' : $st->($tree);
+          ref $tree or return ($tag => {_ => $simple});
           my $node   = $tree->node;
-          my @pairs  = map {$_->($node)} @attrs;
-          defined $simple or @pairs or return ();
-          defined $simple or $simple = 'undef';
-          ($tag => {_ => $simple, @pairs});
+          my @pairs  = map $_->($node), @attrs;
+          defined $simple || @pairs ?  ($tag => {_ => $simple, @pairs}) : ();
         };
 }
 
@@ -650,30 +653,35 @@ sub makeTaggedElement
 #
 
 sub makeMixedElement
-{   my ($self, $path, $tag, $elems, $attrs, $attrs_any) = @_;
+{   my ($self, $path, $tag, $elems, $attrs, $attrs_any,undef,$is_nillable) = @_;
     my @attrs = (odd_elements(@$attrs), @$attrs_any);
     my $mixed = $self->{mixed_elements}
          or panic "how to handle mixed?";
+$is_nillable and panic "nillable mixed not yet supported";
 
       ref $mixed eq 'CODE'
     ? sub { my $tree = shift or return;
             my $node = $tree->node or return;
-            my @v = $mixed->($node);
+            my @v = $mixed->($path, $node);
             @v ? ($tag => $v[0]) : ();
           }
+
     : $mixed eq 'XML_NODE'
-    ? sub { $_[0] ? ($tag => $_[0]->node) : () }
+    ? sub {$_[0] ? ($tag => $_[0]->node) : () }
+
     : $mixed eq 'ATTRIBUTES'
     ? sub { my $tree   = shift or return;
             my $node   = $tree->node;
-            my @pairs  = map {$_->($node)} @attrs;
-            ($tag => {_ => $node, @pairs});
+            my @pairs  = map $_->($node), @attrs;
+            ($tag => { _ => $node, @pairs
+                     , _MIXED_ELEMENT_MODE => 'ATTRIBUTES'});
           } 
     : $mixed eq 'TEXTUAL'
     ? sub { my $tree   = shift or return;
             my $node   = $tree->node;
-            my @pairs  = map {$_->($node)} @attrs;
-            ($tag => {_ => $node->textContent, @pairs});
+            my @pairs  = map $_->($node), @attrs;
+            ($tag => { _ => $node->textContent, @pairs
+                     , _MIXED_ELEMENT_MODE => 'TEXTUAL'});
           } 
     : $mixed eq 'XML_STRING'
     ? sub { my $tree   = shift or return;
@@ -681,7 +689,10 @@ sub makeMixedElement
             ($tag => $node->toString);
           }
     : $mixed eq 'STRUCTURAL'
+
+      # this cannot be reached, because handled somewhere else
     ? panic "mixed structural handled as normal element"
+
     : error __x"unknown mixed_elements value `{value}'", value => $mixed;
 }
 
@@ -690,14 +701,49 @@ sub makeMixedElement
 #
 
 sub makeSimpleElement
-{   my ($self, $path, $tag, $st) = @_;
-    sub { my $value = $st->(@_);
-          defined $value ? ($tag => $value) : ();
-        };
+{   my ( $self, $path, $tag, $st, undef, undef, $comptype, $is_nillable) = @_;
+
+      $is_nillable
+    ? sub { my $tree  = shift or return $st->(undef);
+            my $value = (ref $tree && $tree->nodeNil) ? 'NIL' : $st->($tree);
+            defined $value ? ($tag => $value) : ();
+          }
+    : sub { my $value = $st->(@_);
+            defined $value ? ($tag => $value) : ();
+          };
+
+}
+
+sub default_anytype_handler($$)
+{   my ($path, $node) = @_;
+    ref $node or return $node;
+      (first{ UNIVERSAL::isa($_, 'XML::LibXML::Element') } $node->childNodes)
+    ? $node : $node->textContent;
 }
 
 sub makeBuiltin
 {   my ($self, $path, $node, $type, $def, $check_values) = @_;
+
+    if($type =~ m/}anyType$/)
+    {
+        if(my $a = $self->{any_type})
+        {   return sub {
+               my $node
+                 = ref $_[0] && UNIVERSAL::isa($_[0], 'XML::Compile::Iterator')
+                 ? $_[0]->node : $_[0];
+               $a->( $path, $node, \&default_anytype_handler)};
+        }
+        else
+        {   return sub
+              { ref $_[0] or return $_[0];
+                my $node = UNIVERSAL::isa($_[0], 'XML::Compile::Iterator')
+                  ? $_[0]->node : $_[0];
+                (first{ UNIVERSAL::isa($_, 'XML::LibXML::Element') }
+                     $node->childNodes) ? $node : $node->textContent;
+              };
+        }
+    }
+
     my $check = $check_values ? $def->{check} : undef;
     my $parse = $def->{parse};
     my $err   = $path eq $type
@@ -728,8 +774,6 @@ sub makeBuiltin
       );
 }
 
-# simpleType
-
 sub makeList
 {   my ($self, $path, $st) = @_;
     sub { my $tree = shift;
@@ -738,36 +782,47 @@ sub makeList
              = UNIVERSAL::isa($tree, 'XML::LibXML::Node') ? $tree
              : ref $tree ? $tree->node : undef;
           my $v = ref $tree ? $tree->textContent : $tree;
-          my @v = grep {defined} map {$st->($_, $node)} split(" ",$v);
+          my @v = grep defined, map $st->($_, $node), split " ", $v;
           @v ? \@v : undef;
         };
 }
 
 sub makeFacetsList
 {   my ($self, $path, $st, $info, $early, $late) = @_;
+    my @e = grep defined, @$early;
+    my @l = grep defined, @$late;
+
+    # enumeration and pattern are probably rare
+    @e or return sub {
+        my $values = $st->(@_) or return;
+        $_->($values) for @l;
+        $values;
+    };
+
     sub { defined $_[0] or return undef;
-          my $v = $st->(@_);
-          my @v;
-          if(ref $v eq 'ARRAY') { @v = @$v } # no early limits for extension
-          else
-          {   for(@$early) { defined $v or return (); $v = $_->($v) }
-              @v = defined $v ? split(" ",$v) : ();
-          }
-          my @r;
-      EL: for my $e (@v)
-          {   for(@$late) { defined $e or next EL; $e = $_->($e) }
-              push @r, $e if defined $e;
-          }
-          @r ? \@r : ();
-        };
+        my $list = ref $_[0] ? $_[0]->textContent : $_[0];
+        $_->($list) for @e;
+        my $values = $st->($_[0]) or return;
+        $_->($values) for @l;
+        $values;
+    };
 }
 
 sub makeFacets
 {   my ($self, $path, $st, $info, @do) = @_;
+    @do or return $st;
+
+    @do==1 or return sub
+      { defined $_[0] or return undef;
+        my $v = $st->(@_);
+        for(@do) { defined $v or return (); $v = $_->($v) }
+        $v;
+      };
+
+    my $do = shift @do;
     sub { defined $_[0] or return undef;
           my $v = $st->(@_);
-          for(@do) { defined $v or return (); $v = $_->($v) }
-          $v;
+          defined $v ? $do->($v) : ();
         };
 }
 
@@ -777,7 +832,7 @@ sub makeUnion
           for(@types) { my $v = try { $_->($tree) }; $@ or return $v }
           my $text = $tree->textContent;
 
-          substr $text, 20, -1, '...' if length($text) > 73;
+          substr $text, 20, -5, '...' if length($text) > 50;
           error __x"no match for `{text}' in union at {path}"
              , text => $text, path => $path;
         };
@@ -786,7 +841,7 @@ sub makeUnion
 # Attributes
 
 sub makeAttributeRequired
-{   my ($self, $path, $ns, $tag, $do) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
     sub { my $node = $_[0]->getAttributeNodeNS($ns, $tag);
           defined $node
              or error __x"attribute `{name}' is required at {path}"
@@ -794,12 +849,12 @@ sub makeAttributeRequired
 
           defined $node or return ();
           my $value = $do->($node);
-          defined $value ? ($tag => $value) : ();
+          defined $value ? ($label => $value) : ();
         };
 }
 
 sub makeAttributeProhibited
-{   my ($self, $path, $ns, $tag, $do) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
     sub { my $node = $_[0]->getAttributeNodeNS($ns, $tag);
           defined $node or return ();
           error __x"attribute `{name}' is prohibited at {path}"
@@ -809,54 +864,54 @@ sub makeAttributeProhibited
 }
 
 sub makeAttribute
-{   my ($self, $path, $ns, $tag, $do) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
     sub { my $node = $_[0]->getAttributeNodeNS($ns, $tag);
           defined $node or return ();;
           my $val = $do->($node);
-          defined $val ? ($tag => $val) : ();
+          defined $val ? ($label => $val) : ();
         };
 }
 
 sub makeAttributeDefault
-{   my ($self, $path, $ns, $tag, $do, $default) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do, $default) = @_;
 
     my $mode = $self->{default_values};
     $mode eq 'IGNORE'
         and return sub
           { my $node = $_[0]->getAttributeNodeNS($ns, $tag);
-            defined $node ? ($tag => $do->($node)) : () };
+            defined $node ? ($label => $do->($node)) : () };
 
     my $def = $do->($default);
 
     $mode eq 'EXTEND'
         and return sub
           { my $node = $_[0]->getAttributeNodeNS($ns, $tag);
-            ($tag => ($node ? $do->($node) : $def))
+            ($label => ($node ? $do->($node) : $def))
           };
 
     $mode eq 'MINIMAL'
         and return sub
           { my $node = $_[0]->getAttributeNodeNS($ns, $tag);
             my $v = $node ? $do->($node) : $def;
-            !defined $v || $v eq $def ? () : ($tag => $v);
+            !defined $v || $v eq $def ? () : ($label => $v);
           };
 
     error __x"illegal default_values mode `{mode}'", mode => $mode;
 }
 
 sub makeAttributeFixed
-{   my ($self, $path, $ns, $tag, $do, $fixed) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do, $fixed) = @_;
     my $def  = $do->($fixed);
 
     sub { my $node  = $_[0]->getAttributeNodeNS($ns, $tag)
-              or return ($tag => $def);
+              or return ($label => $def);
 
           my $value = $do->($node);
           defined $value && $value eq $def
               or error __x"value of attribute `{tag}' is fixed to `{fixed}', not `{value}' at {path}"
                   , tag => $tag, fixed => $def, value => $value, path => $path;
 
-          ($tag => $def);
+          ($label => $def);
         };
 }
 
@@ -874,10 +929,9 @@ sub makeSubstgroup
 
           my $do   = $do{$type}
               or return;
-
           my @subst = $do->[1]($tree->descend);
           $tree->nextChild;
-          @subst ? ($do->[0] => $subst[1]) : ();   # rewrite
+          @subst ? ($do->[0] => $subst[1]) : ();   # key-rewrite
         }, 'BLOCK';
 }
 
@@ -887,8 +941,8 @@ sub makeAnyAttribute
 {   my ($self, $path, $handler, $yes, $no, $process) = @_;
     return () unless defined $handler;
 
-    my %yes = map { ($_ => 1) } @{$yes || []};
-    my %no  = map { ($_ => 1) } @{$no  || []};
+    my %yes = map +($_ => 1), @{$yes || []};
+    my %no  = map +($_ => 1), @{$no  || []};
 
     # Takes all, before filtering
     my $all =
@@ -927,14 +981,13 @@ sub makeAnyElement
 {   my ($self, $path, $handler, $yes, $no, $process, $min, $max) = @_;
     $handler ||= 'SKIP_ALL';
 
-    my %yes = map { ($_ => 1) } @{$yes || []};
-    my %no  = map { ($_ => 1) } @{$no  || []};
+    my %yes = map +($_ => 1), @{$yes || []};
+    my %no  = map +($_ => 1), @{$no  || []};
 
     # Takes all, before filtering
-    my $any = $max eq 'unbounded' || $max > 1
+    my $any = ($max eq 'unbounded' || $max > 1)
     ? sub
-      {
-          my $tree  = shift or return ();
+      {   my $tree  = shift or return ();
           my $count = 0;
           my %result;
           while(   (my $child = $tree->currentChild)
@@ -945,7 +998,6 @@ sub makeAnyElement
 
               my $k = pack_type $ns, $child->localName;
               push @{$result{$k}}, $child;
-
               $count++;
               $tree->nextChild;
           }
@@ -956,11 +1008,9 @@ sub makeAnyElement
           %result;
       }
     : sub
-      {   my $tree  = shift or return ();
-          my $child = $tree->currentChild
-              or return ();
-
-          my $ns = $child->namespaceURI || '';
+      {   my $tree  = shift               or return ();
+          my $child = $tree->currentChild or return ();
+          my $ns    = $child->namespaceURI || '';
 
           (!keys %yes || $yes{$ns}) && !(keys %no && $no{$ns})
               or return ();
@@ -990,6 +1040,36 @@ sub makeAnyElement
      bless $run, 'ANY';
 }
 
+# xsi:type handling
+
+sub makeXsiTypeSwitch($$$$)
+{   my ($self, $where, $elem, $default_type, $types) = @_;
+
+    sub {
+        my $tree = shift or return;
+        my $node = $tree->node or return;
+        my $type = $node->getAttributeNS(SCHEMA2001i, 'type');
+        my ($alt, $code);
+        if($type)
+        {   my ($pre, $local) = $type =~ /(.*?)\:(.*)/ ? ($1, $2) : ('',$type);
+            $alt  = pack_type $node->lookupNamespaceURI($pre), $local;
+            $code = $types->{$alt}
+                or error __x"specified xsi:type list for `{default}' does not contain `{got}'"
+                     , default => $default_type, got => $type;
+        }
+        else
+        {   ($alt, $code) = ($default_type, $types->{$default_type});
+        }
+
+        my ($t, $d) = $code->($tree);
+        defined $t or return ();
+
+        $d = { _ => $d } if ref $d ne 'HASH';
+        $d->{XSI_TYPE} ||= $alt;
+        ($t, $d);
+    };
+}
+
 # any kind of hook
 
 sub makeHook($$$$$$)
@@ -999,9 +1079,9 @@ sub makeHook($$$$$$)
     return sub { ($_[0]->node->localName => 'SKIPPED') }
         if $replace && grep {$_ eq 'SKIP'} @$replace;
 
-    my @replace = $replace ? map {$self->_decodeReplace($path,$_)} @$replace:();
-    my @before  = $before  ? map {$self->_decodeBefore($path,$_) } @$before :();
-    my @after   = $after   ? map {$self->_decodeAfter($path,$_)  } @$after  :();
+    my @replace = $replace ? map $self->_decodeReplace($path,$_),@$replace : ();
+    my @before  = $before  ? map $self->_decodeBefore($path,$_), @$before  : ();
+    my @after   = $after   ? map $self->_decodeAfter($path,$_),  @$after   : ();
 
     sub
      { my $tree = shift or return ();
@@ -1010,17 +1090,20 @@ sub makeHook($$$$$$)
        {   $xml = $_->($xml, $path);
            defined $xml or return ();
        }
+
+       my $process = sub { $r->($tree->descend($xml)) };
        my @h = @replace
-             ? map {$_->($xml,$self,$path,$tag)} @replace
-             : $r->($tree->descend($xml));
+         ? map $_->($xml, $self, $path, $tag, $process), @replace
+         : $process->();
+
        @h or return ();
-       my $h = @h==1 ? {_ => $h[0]} : $h[1];  # detect simpleType
+       my $h = @h==1 && !ref $h[0] ? {_ => $h[0]} : $h[1];  # detect simpleType
        foreach my $after (@after)
        {   $h = $after->($xml, $h, $path);
            defined $h or return ();
        }
-       $h;
-     }
+       ($tag => $h);
+     };
 }
 
 sub _decodeBefore($$)
@@ -1028,7 +1111,7 @@ sub _decodeBefore($$)
     return $call if ref $call eq 'CODE';
 
       $call eq 'PRINT_PATH' ? sub {print "$_[1]\n"; $_[0] }
-    : error __x"labeled before hook `{call}' undefined for READER",call=>$call;
+    : error __x"labeled before hook `{call}' undefined for READER", call=>$call;
 }
 
 sub _decodeReplace($$)
@@ -1038,36 +1121,75 @@ sub _decodeReplace($$)
     error __x"labeled replace hook `{call}' undefined for READER", call=>$call;
 }
 
+my %after = 
+  ( PRINT_PATH   => sub {print "$_[2]\n"; $_[1] }
+  , INCLUDE_PATH => sub { my $h = $_[1];
+        $h = { _ => $h } if ref $h ne 'HASH';
+        $h->{_PATH} = $_[0];
+        $h;
+    }
+  , XML_NODE     => sub { my $h = $_[1];
+        $h = { _ => $h } if ref $h ne 'HASH';
+        $h->{_XML_NODE} = $_[0];
+        $h;
+    }
+  , ELEMENT_ORDER => sub { my ($xml, $h) = @_;
+        $h = { _ => $h } if ref $h ne 'HASH';
+        my @order = map type_of_node($_)
+          , grep $_->isa('XML::LibXML::Element'), $xml->childNodes;
+        $h->{_ELEMENT_ORDER} = \@order;
+        $h;
+    }
+  , ATTRIBUTE_ORDER => sub { my ($xml, $h) = @_;
+        $h = { _ => $h } if ref $h ne 'HASH';
+        my @order = map $_->nodeName, $xml->attributes;
+        $h->{_ATTRIBUTE_ORDER} = \@order;
+        $h;
+    }
+  , NODE_TYPE => sub { my ($xml, $h) = @_;
+        $h = { _ => $h } if ref $h ne 'HASH';
+        $h->{_NODE_TYPE} = type_of_node $xml;
+        $h;
+    }
+  );
+
 sub _decodeAfter($$)
 {   my ($self, $path, $call) = @_;
     return $call if ref $call eq 'CODE';
 
-      $call eq 'PRINT_PATH' ? sub {print "$_[2]\n"; $_[1] }
-    : $call eq 'XML_NODE'  ?
-      sub { my $h = $_[1];
-            $h = { _ => $h } if ref $h ne 'HASH';
-            $h->{_XML_NODE} = $_[0];
-            $h;
-          }
-    : $call eq 'ELEMENT_ORDER' ?
-      sub { my ($xml, $h) = @_;
-            $h = { _ => $h } if ref $h ne 'HASH';
-            my @order = map {type_of_node $_}
-                grep { $_->isa('XML::LibXML::Element') }
-                    $xml->childNodes;
-            $h->{_ELEMENT_ORDER} = \@order;
-            $h;
-          }
-    : $call eq 'ATTRIBUTE_ORDER' ?
-      sub { my ($xml, $h) = @_;
-            $h = { _ => $h } if ref $h ne 'HASH';
-            my @order = map {$_->nodeName} $xml->attributes;
-            $h->{_ATTRIBUTE_ORDER} = \@order;
-            $h;
-          }
-    : error __x"labeled after hook `{call}' undefined for READER", call=> $call;
+    # The 'after' can be called on a single.  In that case, turn it into
+    # a HASH for additional information.
+    my $dec = $after{$call}
+        or error __x"labeled after hook `{call}' undefined for READER"
+            , call=> $call;
+
+    $dec;
 }
 
+sub makeBlocked($$$)
+{   my ($self, $where, $class, $type) = @_;
+
+    # errors are produced in class=misfit to allow other choices to succeed.
+      $class eq 'anyType'
+    ? { st => sub { error __x"use of `{type}' blocked at {where}"
+              , type => $type, where => $where, _class => 'misfit';
+          }}
+    : $class eq 'simpleType'
+    ? { st => sub { error __x"use of {class} `{type}' blocked at {where}"
+              , class => $class, type => $type, where => $where
+              , _class => 'misfit';
+          }}
+    : $class eq 'complexType'
+    ? { elems => [] }
+    : $class eq 'ref'
+    ? { st => sub { error __x"use of referenced `{type}' blocked at {where}"
+              , type => $type, where => $where, _class => 'misfit';
+          }}
+    : panic "blocking of $class for $type not implemented";
+}
+
+#-----------------------------------
+
 =chapter DETAILS
 
 =section Processing Wildcards
@@ -1179,9 +1301,22 @@ from the container which are not used.  Also in this case C<TAKE_ALL>
 is required to produce C<any> results.  C<SKIP_ALL> will ignore all
 results, although this are being processed for validation needs.
 
-The C<minOccurs> and C<maxOccurs> of C<any> are ignored: the amount of
-elements is always unbounded.  Therefore, you will get an array of
-elements back per type. 
+=subsection option any_type CODE
+
+By default, the elements which have type "xsd:anyType" will return
+an M<XML::LibXML::Element> when there are sub-elements.  Otherwise,
+it will return the textual content. 
+
+If you pass your own CODE reference, you can change this behavior.  It
+will get called with the path, the node, and the default handler.  Be
+awayre the $node may actually be a string already.
+
+   $schema->compile(READER => ..., any_type => \&handle_any_type);
+   sub handle_any_type($$$)
+   { my ($path, $node, $handler) = @_;
+     ref $node or return $node;
+     $node;
+   }
 
 =section Mixed elements
 
@@ -1199,6 +1334,11 @@ is needed as text, or the element should be parsed and the text ignored.
 The reader has various options to avoid the need of processing raw
 XML::LibXML nodes.
 
+[1.00]
+When the return is a HASH, that HASH will also contain the
+C<_MIXED_ELEMENT_MODE> key, to help people understand what
+happens.  This is not possible for all modes, only for some.
+
 With M<XML::Compile::Schema::compile(mixed_elements)> set to
 =over 4
 
@@ -1252,13 +1392,17 @@ is returned, the whole node will disappear.
 This hook offers a predefined C<PRINT_PATH>.
 
 =example to trace the paths
- $schema->addHook(path => qr/./, before => 'PRINT_PATH');
+ $schema->addHook
+   ( action => 'READER'
+   , path   => qr/./
+   , before => 'PRINT_PATH'
+   );
 
 =subsection hooks executed as replacement
 
-Your C<replace> hook should return a list of key-value pairs. To
-produce it, it will get the M<XML::LibXML::Element>, the translator settings
-as HASH, the path, and the localname.
+Your C<replace> hook should return a list of key-value pairs. To produce
+it, it will get the M<XML::LibXML::Element>, the translator settings as
+HASH, the path, and the localname.
 
 This hook has a predefined C<SKIP>, which will not process the
 found element, but simply return the string "SKIPPED" as value.
@@ -1272,25 +1416,28 @@ M<XML::LibXML::Simple> to translate a part of your tree.  Simply
 
  use XML::LibXML::Simple  qw/XMLin/;
  $schema->addHook
-   ( type    => ...bad-type-definition...
+   ( action  => 'READER'
+   , type    => 'tns:xyz'     # or pack_type($tns,'xyz')
+  #  path    => qr!/company$! # by element name
    , replace =>
-       sub { my ($xml, $args, $path, $type) = @_;
+       sub { my ($xml, $args, $path, $type, $r) = @_;
              ($type => XMLin($xml, ...));
            }
    );
 
 =subsection hooks for post-processing, after the data is collected
 
-The data is collect, and passed as second argument after the XML node.
-The third argument is the path.  Be careful that the collected data
+Your code reference gets called with three parameters: the XML node,
+the data collected and the path.  Be careful that the collected data
 might be a SCALAR (for simpleType).  Return a HASH or a SCALAR.  C<undef>
 may work, unless it is the value of a required element you throw awy.
 
 This hook also offers a predefined C<PRINT_PATH>.  Besides, it
-has C<XML_NODE>, C<ELEMENT_ORDER>, and C<ATTRIBUTE_ORDER>, which will
-result in additional fields in the HASH, respectively containing the
-CODE which was processed, the element names, and the attribute names.
-The keys start with an underscore C<_>.
+has C<INCLUDE_PATH>, C<XML_NODE>, C<NODE_TYPE>, C<ELEMENT_ORDER>,
+and C<ATTRIBUTE_ORDER>, which will result in additional fields in
+the HASH, respectively containing the NODE which was processed (an
+XML::LibXML::Element), the type_of_node, the element names, and the
+attribute names.  The keys start with an underscore C<_>.
 
 =section Typemaps
 
@@ -1303,7 +1450,7 @@ differently.  This section is about how the reader handles typemaps.
 Usually, an XML type will be mapped on a Perl class.  The Perl class
 implements the C<fromXML> method as constructor.
 
- $schema->typemap($sometype => 'My::Perl::Class');
+ $schema->addTypemaps($sometype => 'My::Perl::Class');
 
  package My::Perl::Class;
  ...
@@ -1333,7 +1480,7 @@ the XML structure:
 
 =subsection Typemap to Object
 
-An other option is to implement an object factory: one object which creates
+Another option is to implement an object factory: one object which creates
 other objects.  In this case, the C<$xmltype> parameter can come of use,
 to have one object spawning many different other objects.
 
