@@ -2,14 +2,17 @@
 package XML::Compile::Translate::Template;
 use base 'XML::Compile::Translate';
 
-use XML::Compile::Translate::Writer;
-
 use strict;
 use warnings;
-no warnings 'once';
+no warnings 'once', 'recursion';
 
-use XML::Compile::Util qw/odd_elements unpack_type/;
+use XML::Compile::Util
+  qw/odd_elements even_elements SCHEMA2001i pack_type unpack_type/;
 use Log::Report 'xml-compile', syntax => 'SHORT';
+use List::Util  qw/max/;
+
+our $VERSION;         # OODoc adds $VERSION to the script
+$VERSION ||= 'undef';
 
 =chapter NAME
 
@@ -37,21 +40,68 @@ the schema describes.
 
 =cut
 
-BEGIN {
-   no strict 'refs';
-   *$_ = *{"XML::Compile::Translate::Writer::$_"}
-      for qw/makeTagQualified makeTagUnqualified/;
+sub makeTagQualified
+{   my ($self, $path, $node, $local, $ns) = @_;
+    my $prefix = $self->_registerNSprefix('', $ns, 1);
+
+    # it is certainly not correct to do a keyRewrite here, but it works :(
+      $self->{_output} eq 'PERL' ? $self->keyRewrite($ns, $local)
+    : length $prefix             ? "$prefix:$local"
+    :                              $local;
 }
 
-sub actsAs($) { $_[1] eq 'READER' }
+sub makeTagUnqualified
+{   my ($self, $path, $node, $local, $ns) = @_;
+#   $name =~ s/.*\://;
+    return $self->keyRewrite($ns, $local)
+        if $self->{_output} eq 'PERL';
 
-sub makeWrapperNs
-{   my ($self, $path, $processor, $index) = @_;
-    $processor;
+    my $prefix = $self->_registerNSprefix('', $ns, 1);
+    length $prefix ? "$prefix:$local" : $local;
+}
+
+# Detect recursion.  Based on type is best, but some schema's do not
+# have named types, so tags are indexed as well.
+my (%recurse_type, %reuse_type, %recurse_tag, %reuse_tag);
+
+sub compile($@)
+{   my ($self, $type, %args) = @_;
+    $self->{_output} = $args{output};
+    (%recurse_type, %reuse_type, %recurse_tag, %reuse_tag) = ();
+    $self->SUPER::compile($type, %args);
+}
+
+sub actsAs($)
+{   my ($self, $as) = @_;
+       ($as eq 'READER' && $self->{_output} eq 'PERL')
+    || ($as eq 'WRITER' && $self->{_output} eq 'XML')
+}
+
+sub makeWrapperNs($$$$$)
+{   my ($self, $path, $processor, $index, $filter) = @_;
+
+    my @entries;
+    $filter = sub {1} if ref $filter ne 'CODE';
+
+    foreach my $entry (sort {$a->{prefix} cmp $b->{prefix}} values %$index)
+    {   $entry->{used} or next;
+        my ($prefix, $uri) = @{$entry}{'prefix', 'uri'};
+        $filter->($uri, $prefix) or next;
+        push @entries, [ $uri, $prefix ];
+        $entry->{used} = 0;
+    }
+
+    sub { my $data = $processor->(@_) or return ();
+          if($self->{include_namespaces})
+          {   $data->{"xmlns:$_->[1]"} = $_->[0] for @entries;
+          }
+          $data;
+        };
 }
 
 sub typemapToHooks($$)
 {   my ($self, $hooks, $typemap) = @_;
+
     while(my($type, $action) = each %$typemap)
     {   defined $action or next;
 
@@ -85,17 +135,20 @@ sub makeElementWrapper
 sub _block($@)
 {   my ($self, $block, $path, @pairs) = @_;
     bless
-    sub { my @elems  = map { $_->() } odd_elements @pairs;
+    sub { my @elems  = map { $_->()    } odd_elements @pairs;
           my @tags   = map { $_->{tag} } @elems;
 
           local $" = ', ';
-          my $struct = @tags ? "$block of @tags" : "empty $block";
+          my $struct = @tags ? "$block of @tags"
+            : "empty $block from ".join(" ", even_elements @pairs);
+
           my @lines;
           while(length $struct > 65)
-          {   $struct =~ s/(.{1,60})(\s)//;
+          {   $struct =~ s/(.{1,60}|\S+)(?:\s+|$)//;
               push @lines, $1;
           }
-          push @lines, $struct;
+          push @lines, $struct
+              if length $struct;
           $lines[$_] =~ s/^/  / for 1..$#lines;
 
            { tag    => $block
@@ -120,7 +173,7 @@ sub makeBlockHandler
            : $max ne 'unbounded' && $max==1 && $min==1 ? ''  # the usual case
            :       "occurs $min <= # <= $max times";
 
-          $data->{occur}   = $occur if $occur;
+          $data->{occur} ||= $occur if $occur;
           if($max ne 'unbounded' && $max==1)
           {   bless $data, 'BLOCK';
           }
@@ -142,7 +195,7 @@ sub makeElementHandler
            : $max ne 'unbounded' && $max==1 && $min==0 ? 'is optional' 
            : $max ne 'unbounded' && $max==1 && $min==1 ? ''  # the usual case
            :                                  "occurs $min <= # <= $max times";
-          $data->{occur}    = $occur if $occur;
+          $data->{occur}  ||= $occur if $occur;
           $data->{is_array} = $max eq 'unbounded' || $max > 1;
           $data;
         };
@@ -160,96 +213,129 @@ sub makeElementHref
 
 sub makeElement
 {   my ($self, $path, $ns, $childname, $do) = @_;
-    $do;
+    sub { my $h = $do->(@_);
+          $h->{_NAME} = $childname;
+          $h;
+    };
 }
 
 sub makeElementDefault
 {   my ($self, $path, $ns, $childname, $do, $default) = @_;
-    sub { (occur => "$childname defaults to example $default",  $do->()) };
+    sub { my $h = $do->(@_);
+          $h->{occur}   = "defaults to '$default'";
+          $h->{example} = $default;
+          $h;
+        };
 }
 
 sub makeElementFixed
 {   my ($self, $path, $ns, $childname, $do, $fixed) = @_;
-    $fixed   = $fixed->example;
-    sub { (occur => "$childname fixed to example $fixed", $do->()) };
-}
-
-sub makeElementNillable
-{   my ($self, $path, $ns, $childname, $do) = @_;
-    sub { (occur => "$childname is nillable", $do->()) };
+    sub { my $h = $do->(@_);
+          $h->{occur}   = "fixed to '$fixed'";
+          $h->{example} = $fixed;
+          $h;
+        };
 }
 
 sub makeElementAbstract
 {   my ($self, $path, $ns, $childname, $do) = @_;
-    sub { () };
+#   sub { () };
+    sub {
+       my $h = $do->(@_);
+       $h->{_NAME} = $childname;
+       $h->{occur} = "ABSTRACT";
+       $h;
+    };
 }
 
-my %recurse;
 sub makeComplexElement
-{   my ($self, $path, $tag, $elems, $attrs, $any_attr) = @_;
+{   my ($self, $path, $tag, $elems, $attrs, $any_attr, $type, $is_nillable)=@_;
     my @parts = (odd_elements(@$elems, @$attrs), @$any_attr);
 
     sub { my (@attrs, @elems);
+          my $is_pseudo_type = $type !~ m/^{/;  # like "unnamed complex"
 
-          if($recurse{$tag})
+          if((!$is_pseudo_type && $recurse_type{$type}) || $recurse_tag{$tag})
           {   return
               +{ kind   => 'complex'
                , struct => 'probably a recursive complex'
                , tag    => $tag
+               , _TYPE  => $type
                };
           }
 
-          $recurse{$tag}++;
+          if((!$is_pseudo_type && $reuse_type{$type}) || $reuse_tag{$tag})
+          {   return
+              +{ kind   => 'complex'
+               , struct => 'complex structure shown above'
+               , tag    => $tag
+               , _TYPE  => $type
+               };
+          }
+
+          $recurse_type{$type}++; $recurse_tag{$tag}++;
+          $reuse_type{$type}++;   $reuse_tag{$tag}++;
           foreach my $part (@parts)
           {   my $child = $part->();
               if($child->{attr}) { push @attrs, $child }
               else               { push @elems, $child }
           }
-          $recurse{$tag}--;
+          $recurse_type{$type}--; $recurse_tag{$tag}--;
 
-          +{ kind    => 'complex'
-#          , struct  => "$tag is complex"  # too obvious to mention
-           , tag     => $tag
-           , attrs   => \@attrs
-           , elems   => \@elems
+          +{ kind   => 'complex'
+           , struct => ($is_nillable ? "is nillable, as: $tag => NIL" : undef)
+           , tag    => $tag
+           , attrs  => \@attrs
+           , elems  => \@elems
+           , _TYPE  => $type
            };
         };
 }
 
 sub makeTaggedElement
-{   my ($self, $path, $tag, $st, $attrs, $attrs_any) = @_;
+{   my ($self, $path, $tag, $st, $attrs, $attrs_any, $type, $is_nillable) = @_;
     my @parts = (odd_elements(@$attrs), @$attrs_any);
 
-    my %content =
-     ( tag     => '_'
-     , struct  => 'string content of the container'
-     , example => 'Hello, World!' 
-     );
-
     sub { my @attrs  = map {$_->()} @parts;
-          my $simple = $st->() || '';
+          my %simple = $st->();
+
+          my @struct = 'string content of the container';
+          push @struct, $simple{struct} if $simple{struct};
+          push @struct, 'is nillable, hence value or NIL' if $is_nillable;
+
+          my %content =
+            ( tag     => '_'
+            , struct  => \@struct
+            , example => ($simple{example} || 'Hello, World!')
+            );
+          $content{_TYPE}   = $simple{_TYPE}   if $simple{_TYPE};
 
           +{ kind    => 'tagged'
            , struct  => "$tag is simple value with attributes"
            , tag     => $tag
            , attrs   => \@attrs
            , elems   => [ \%content ]
+           , _TYPE   => $type
            };
         };
 }
 
 sub makeMixedElement
-{   my ($self, $path, $tag, $elems, $attrs, $attrs_any) = @_;
+{   my ($self, $path, $tag, $elems, $attrs, $attrs_any, $type, $is_nillable)=@_;
     my @parts = (odd_elements(@$attrs), @$attrs_any);
 
+    my @struct = 'mixed content cannot be processed automatically';
+    push @struct, 'is nillable' if $is_nillable;
+
     my %mixed =
      ( tag     => '_'
-     , struct  => "mixed content cannot be processed automatically"
+     , struct  => \@struct
      , example => "XML::LibXML::Element->new('$tag')"
      );
 
     unless(@parts)   # show simpler alternative
-    {   $mixed{tag} = $tag;
+    {   $mixed{tag}   = $tag;
+        $mixed{type} = $type;
         return sub { \%mixed };
     }
 
@@ -259,14 +345,17 @@ sub makeMixedElement
            , tag     => $tag
            , elems   => [ \%mixed ]
            , attrs   => \@attrs
+           , _TYPE   => $type
            };
         };
 }
 
 sub makeSimpleElement
-{   my ($self, $path, $tag, $st) = @_;
-    sub { +{ kind    => 'simple'
-#          , struct  => "elem $tag is a single value"  # too obvious
+{   my ($self, $path, $tag, $st, undef, undef, $type, $is_nillable) = @_;
+    sub { my @struct;
+          push @struct, 'is nillable, hence value or NIL' if $is_nillable;
+          +{ kind    => 'simple'
+           , struct  => \@struct
            , tag     => $tag
            , $st->()
            };
@@ -275,41 +364,67 @@ sub makeSimpleElement
 
 sub makeBuiltin
 {   my ($self, $path, $node, $type, $def, $check_values) = @_;
-    my $example = $def->{example};
-    sub { (type => $type, example => $example) };
+    sub { (_TYPE=> $type, example => $def->{example}) };
 }
 
 sub makeList
 {   my ($self, $path, $st) = @_;
-    sub { (struct => "a (blank separated) list of elements", $st->()) };
+    sub { my %d = $st->();
+          $d{struct} = 'a list of values, where each';
+          my $example = $d{example};
+          $example    = qq("$example") if $example =~ m/[^0-9.]/;
+          $d{example} = "[ $example , ... ]";
+          %d };
 }
 
 sub makeFacetsList
-{   my ($self, $path, $st, $info, $early, $late) = @_;
-    sub { (facets => "with some restrictions on list elements", $st->()) };
+{   my ($self, $path, $st, $info) = @_;
+    $self->makeFacets($path, $st, $info);
 }
 
-sub _fillFacets($@)
-{   my ($self,$type) = (shift, shift);
+sub _ff($@)
+{  my ($self,$type) = (shift, shift);
     my @lines = $type.':';
     while(@_)
     {   my $facet = shift;
+        $facet =~ s/\t/\t/g;
+        $facet = qq{"$facet"} if $facet =~ m/\s/;
         push @lines, '  ' if length($lines[-1]) + length($facet) > 55;
         $lines[-1] .= ' '.$facet;
     }
-    \@lines;
+    @lines;
 }
 
 sub makeFacets
-{   my ($self, $path, $st, $info, @do) = @_;
-    my $comment
-       = keys %$info==1 && $info->{enumeration}
-       ? $self->_fillFacets('Enum', sort @{$info->{enumeration}})
-       : keys %$info==1 && exists $info->{pattern}
-       ? $self->_fillFacets('Pattern', @{$info->{pattern}})
-       : "with some value restrictions";
+{   my ($self, $path, $st, $info) = @_;
+    my @comment;
+    foreach my $k (sort keys %$info)
+    {   my $v = $info->{$k};
+        push @comment
+        , $k eq 'enumeration'  ? $self->_ff('Enum', sort @$v)
+        : $k eq 'pattern'      ? $self->_ff('Pattern', @$v)
+        : $k eq 'length'       ? "fixed length of $v"
+        : $k eq 'maxLength'    ? "length <= $v"
+        : $k eq 'minLength'    ? "length >= $v"
+        : $k eq 'totalDigits'  ? "total digits is $v"
+        : $k eq 'maxScale'     ? "scale <= $v"
+        : $k eq 'minScale'     ? "scale >= $v"
+        : $k eq 'maxInclusive' ? "value <= $v"
+        : $k eq 'maxExclusive' ? "value < $v"
+        : $k eq 'minInclusive' ? "value >= $v"
+        : $k eq 'minExclusive' ? "value >  $v"
+        : $k eq 'fractionDigits' ? "faction digits is $v"
+        : $k eq 'whiteSpace'   ? "white-space $v"
+        : "restriction? $k = $v";
+    }
 
-    sub { (facets => $comment, $st->()) };
+    my %facet = (facets => \@comment, $st->());
+
+    if(my $enum = $info->{enumeration})
+    {   $facet{example} = $enum->[0];
+    }
+
+    sub { %facet };
 }
 
 sub makeUnion
@@ -324,47 +439,47 @@ sub makeUnion
 }
 
 sub makeAttributeRequired
-{   my ($self, $path, $ns, $tag, $do) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
 
-    sub { +{ kind    => 'attr'
-           , tag     => $tag
-           , occurs  => "attribute $tag is required"
+    sub { +{ kind   => 'attr'
+           , tag    => $label
+           , occur  => "attribute $tag is required"
            , $do->()
            };
         };
 }
 
 sub makeAttributeProhibited
-{   my ($self, $path, $ns, $tag, $do) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
     ();
 }
 
 sub makeAttribute
-{   my ($self, $path, $ns, $tag, $do) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
     sub { +{ kind    => 'attr'
-           , tag     => $tag
+           , tag     => $label
            , $do->()
            };
         };
 }
 
 sub makeAttributeDefault
-{   my ($self, $path, $ns, $tag, $do) = @_;
-    sub { +{ kind   => 'attr'
-           , tag    => $tag
-           , occurs => "attribute $tag has default"
+{   my ($self, $path, $ns, $tag, $label, $do) = @_;
+    sub { +{ kind  => 'attr'
+           , tag   => $label
+           , occur => "attribute $tag has default"
            , $do->()
            };
         };
 }
 
 sub makeAttributeFixed
-{   my ($self, $path, $ns, $tag, $do, $fixed) = @_;
+{   my ($self, $path, $ns, $tag, $label, $do, $fixed) = @_;
     my $value = $fixed->value;
 
-    sub { +{ kind    => 'attr'
-           , tag     => $tag
-           , occurs  => "attribute $tag is fixed"
+    sub { +{ kind   => 'attr'
+           , tag    => $label
+           , occur  => "attribute $tag is fixed"
            , example => $value
            };
         };
@@ -372,12 +487,61 @@ sub makeAttributeFixed
 
 sub makeSubstgroup
 {   my ($self, $path, $type, @do) = @_;
-    my @tags = sort map { $_->[0] } odd_elements @do;
 
-    sub { +{ kind    => 'substitution group'
-           , tag     => $do[1][0]
-           , struct  => [ "substitutionGroup $type:", map { "   $_" } @tags ]
-           , example => "{ $tags[0] => {...} }"
+    sub {
+        my (@example_tags, $example_nest, %tags);
+        my $group = $do[1][0];
+
+        while(@do)
+        {   my ($type,  $info) = (shift @do, shift @do);
+            my ($label, $call) = @$info;
+            my $processed = $call->();
+            my $show = '';
+            if(my $type = $processed->{_TYPE})
+            {   $show = $self->prefixed($type);
+            }
+
+            if($processed->{occur} && $processed->{occur} eq 'ABSTRACT')
+            {   $show .= ' (abstract)';
+            }
+            else
+            {   # some complication to always produce the same tag for
+                # regression tests... Instance uses a HASH...
+                push @example_tags, $label;
+                $example_nest ||= $processed->{kind} eq 'simple'
+                    ? ($processed->{example} || '...') : '{...}';
+            }
+        
+            $tags{$label} = $show;
+        }
+
+        my $longest = max map length, keys %tags;
+        my @lines = map sprintf("  %-${longest}s %s", $_, $tags{$_}),
+            sort keys %tags;
+
+        my $example_tag = (sort @example_tags)[0];
+        my $example = $example_tag ? "{ $example_tag => $example_nest }"
+          : "undef  # only abstract types known";
+
+        my $name    = $self->prefixed($type);
+
+       +{ kind    => 'substitution group'
+        , tag     => $group
+        , struct  => [ "substitutionGroup $name", @lines ]
+        , example => $example
+        }
+    };
+}
+
+sub makeXsiTypeSwitch($$$$)
+{   my ($self, $where, $elem, $default_type, $types) = @_;
+    my @types   = map "  ".$self->prefixed($_), sort keys %$types;
+    my $deftype = $self->prefixed($default_type);
+
+    sub { +{ kind    => 'xsi:type switch'
+           , tag     => $elem
+           , struct  => [ 'xsi:type alternatives:', @types ]
+           , example => "{ XSI_TYPE => '$deftype', %data }"
            }
         };
 }
@@ -385,17 +549,31 @@ sub makeSubstgroup
 sub makeAnyAttribute
 {   my ($self, $path, $handler, $yes, $no, $process) = @_;
     $yes ||= []; $no ||= [];
-    my $occurs = @$yes ? "in @$yes" : @$no ? "not in @$no" : 'any type';
-    bless sub { +{kind => 'attr' , struct  => "anyAttribute $occurs"
+    $yes = [ map {$self->prefixed("{$_}") || $_} @$yes];
+    $no  = [ map {$self->prefixed("{$_}") || $_} @$no];
+    my $occurs = @$yes ? "in @$yes" : @$no ? "not in @$no" : 'in any namespace';
+    bless sub { +{kind => 'attr' , struct  => "any attribute $occurs"
                  , tag => 'ANYATTR', example => 'AnySimple'} }, 'ANY';
 }
 
 sub makeAnyElement
 {   my ($self, $path, $handler, $yes, $no, $process, $min, $max) = @_;
     $yes ||= []; $no ||= [];
-    my $occurs = @$yes ? "in @$yes" : @$no ? "not in @$no" : 'any type';
-    bless sub { +{ kind => 'element', struct  => 'anyElement'
-                 , tag => "ANY", example => 'ANY' } }, 'ANY';
+    $yes = [ map {$self->prefixed("{$_}") || $_} @$yes];
+    $no  = [ map {$self->prefixed("{$_}") || $_} @$no];
+    my $where = @$yes ? "in @$yes" : @$no ? "not in @$no" : 'in any namespace';
+
+    my $data  = +{ kind => 'element', struct  => "any element $where"
+                 , tag => "ANY", example => 'Anything' };
+    my $occur
+      = $max eq 'unbounded' && $min==0 ? 'occurs any number of times'
+      : $max ne 'unbounded' && $max==1 && $min==0 ? 'is optional' 
+      : $max ne 'unbounded' && $max==1 && $min==1 ? ''  # the usual case
+      :                                  "occurs $min <= # <= $max times";
+    $data->{occur}  ||= $occur if $occur;
+    $data->{is_array} = $max eq 'unbounded' || $max > 1;
+
+    bless sub { +$data }, 'ANY';
 }
 
 sub makeHook($$$$$$)
@@ -413,13 +591,14 @@ sub makeHook($$$$$$)
     my @after   = $after   ? map {$self->_decodeAfter($path,$_)  } @$after  :();
 
     sub
-    {  for(@before) { $_->($tag, $path) or return }
+    {   my $doc = XML::LibXML::Document->new;
+        for(@before) { $_->($doc, $path, undef) or return }
 
-       my $d = @replace ? $replace[0]->($tag, $path) : $r->();
-       defined $d or return ();
+       my $xml = @replace ? $replace[0]->($doc, $path, $r) : $r->();
+       defined $xml or return ();
 
-       for(@after) { $d = $_->($d, $tag, $path) or return }
-       $d;
+       for(@after) { $xml = $_->($doc, $path, $xml) or return }
+       $xml;
      }
 }
 
@@ -433,6 +612,18 @@ sub _decodeReplace($$)
 {   my ($self, $path, $call) = @_;
     return $call if ref $call eq 'CODE';
 
+    if($call eq 'COLLAPSE')
+    {   return sub 
+         {  my ($tag, $path, $do) = @_;
+panic "$tag, $path, $do\n" unless ref $do;
+            my $h = $do->();
+            $h->{elems} = [ { struct => [ 'content collapsed' ]
+                            , kind   => 'collapsed' } ];
+            delete $h->{attrs};
+            $h;
+         };
+    }
+
     # SKIP already handled
     error __x"labeled replace hook `{name}' undefined", name => $call;
 }
@@ -450,7 +641,35 @@ sub _decodeAfter($$)
 
 sub toPerl($%)
 {   my ($self, $ast, %args) = @_;
-    my @lines = $self->_perlAny($ast, \%args);
+    $ast or return undef;
+
+    my @lines;
+    if($ast->{kind})
+    {   my $name = $ast->{_NAME} || $ast->{tag};
+        my $pref = $self->prefixed($name);
+        push @lines, defined $pref
+          ? ("# Describing $ast->{kind} $pref", "#     $name")
+          :  "# Describing $ast->{kind} $name";
+    }
+
+    push @lines
+      , "#"
+      , "# Produced by ".__PACKAGE__." version $VERSION"
+      , "#          on ".localtime()
+      , "#"
+      , "# BE WARNED: in most cases, the example below cannot be used without"
+      , "# interpretation.  The comments will guide you."
+      , "#"
+        unless $args{skip_header};
+
+    # add info about name-spaces
+    foreach my $nsdecl (grep /^xmlns\:/, sort keys %$ast)
+    {   push @lines, sprintf "# %-15s %s", $nsdecl, $ast->{$nsdecl} || '(none)';
+    }
+    push @lines, '' if @lines;
+    
+    # produce data tree
+    push @lines, $self->_perlAny($ast, \%args);
 
     # remove leading  'type =>'
     for(my $linenr = 0; $linenr < @lines; $linenr++)
@@ -465,21 +684,32 @@ sub toPerl($%)
     $lines;
 }
 
+my %seen;
 sub _perlAny($$);
 sub _perlAny($$)
 {   my ($self, $ast, $args) = @_;
 
     my @lines;
+    if($ast->{_TYPE} && $args->{show_type})
+    {   my $pref = $self->prefixed($ast->{_TYPE});
+        if($pref)
+        {   push @lines  # not perfect, but a good attempt
+              , $pref =~ m/^[aiou]/i && $pref !~ m/^(uni|eu)/i
+              ? "# is an $pref" : "# is a $pref";
+        }
+    }
+
     if($ast->{struct} && $args->{show_struct})
     {   my $struct = $ast->{struct};
         my @struct = ref $struct ? @$struct : $struct;
         s/^/# /gm for @struct;
         push @lines, @struct;
     }
-    push @lines, "# is a $ast->{type}" if $ast->{type} && $args->{show_type};
-    push @lines, "# $ast->{occur}"  if $ast->{occur}   && $args->{show_occur};
 
-    if($ast->{facets}  && $args->{show_facets})
+    push @lines, "# $ast->{occur}"
+        if $ast->{occur} && $args->{show_occur};
+
+    if($ast->{facets} && $args->{show_facets})
     {   my $facets = $ast->{facets};
         my @facets = ref $facets ? @$facets : $facets;
         s/^/# /gm for @facets;
@@ -497,14 +727,17 @@ sub _perlAny($$)
         @sub or next;
 
         # last line is code and gets comma
-        $sub[-1] =~ s/\,?\s*$/,/;
+        $sub[-1] =~ s/\,?\s*$/,/
+            if $sub[-1] !~ m/\#\s/;
 
         if(ref $ast ne 'BLOCK')
         {   s/^(.)/$args->{indent}$1/ for @sub;
         }
 
         # seperator blank, sometimes
-        unshift @sub, '' if $sub[0] =~ m/^\s*[#{]/;  # } 
+        unshift @sub, ''
+            if $sub[0] =~ m/^\s*[#{]/   # } 
+            || (@subs && $subs[-1] =~ m/[}\]]\,\s*$/);
 
         push @subs, @sub;
     }
@@ -512,8 +745,9 @@ sub _perlAny($$)
     if(ref $ast eq 'REP-BLOCK')
     {  # repeated block
        @subs or @subs = '';
-       $subs[0]  =~ s/^  /{ /;
-       $subs[-1] =~ s/$/ },/;
+       $subs[0] =~ s/^  /{ / or $subs[0] =~ s/^\s*$/{/;
+       if($subs[-1] =~ m/\#\s/) { push @subs, "}," }
+       else { $subs[-1] =~ s/$/ },/ }
     }
 
     # XML does not permit difficult tags, but we still check.
@@ -538,18 +772,22 @@ sub _perlAny($$)
         if($ast->{is_array})
         {   s/^(.)/  $1/ for @subs;
             $subs[0]  =~ s/^[ ]{0,3}/[ {/;
-            $subs[-1] =~ s/$/ }, ], /;
+            if($subs[-1] =~ m/\#\s/) { push @subs, "}, ], " }
+            else {$subs[-1] .= ' }, ], ' }
             push @lines, "$tag =>", @subs;
         }
         else
         {   $subs[0]  =~ s/^  /{ /;
-            $subs[-1] =~ s/$/ },/;
+            if($subs[-1] =~ m/\#\s/) { push @subs, "}, " }
+            else {$subs[-1] .= ' },' }
             push @lines, "$tag =>", @subs;
         }
     }
     elsif($kind eq 'complex' || $kind eq 'mixed')  # empty complex-type
-    {   push @lines, "$tag => {}";
+    {   # if there is an "occurs", then there can always be more than one
+        push @lines, $tag.' => '.($ast->{occur} ? '[{},]' : '{}');
     }
+    elsif($kind eq 'collapsed') {;}
     elsif($kind eq 'union')    # union type
     {   foreach my $union ( @{$ast->{choice}} )
         {  # remove examples
@@ -559,19 +797,22 @@ sub _perlAny($$)
            push @lines, @l;
         }
     }
-    elsif(!$ast->{example})
+    elsif(!exists $ast->{example})
     {   push @lines, "$tag => 'TEMPLATE-ERROR $ast->{kind}'";
     }
 
-    if(my $example = $ast->{example})
+    my $example = $ast->{example};
+    if(defined $example)
     {   $example = qq{"$example"}      # in quotes unless
           if $example !~ m/^[+-]?\d+(?:\.\d+)?$/  # numeric or
           && $example !~ m/^\$/                   # variable or
           && $example !~ m/^bless\b/              # constructor or
-          && $example !~ m/^\$?[\w:]*\-\>/;       # method call example
+          && $example !~ m/^\$?[\w:]*\-\>/        # method call example
+          && $example !~ m/^\{.*\}$/              # anon HASH example
+          && $example !~ m/^\[.*\]$/;             # anon ARRAY example
 
         push @lines, "$tag => "
-          . ($ast->{is_array} ? " [ $example, ]" : $example);
+          . ($ast->{is_array} ? "[ $example, ]" : $example);
     }
     @lines;
 }
@@ -582,13 +823,46 @@ sub _perlAny($$)
 
 sub toXML($$%)
 {   my ($self, $doc, $ast, %args) = @_;
-    $self->_xmlAny($doc, $ast, "\n$args{indent}", \%args);
+    my $xml = $self->_xmlAny($doc, $ast, "\n$args{indent}", \%args);
+
+    UNIVERSAL::isa($xml, 'XML::LibXML::Element')
+        or return $xml;
+
+    # add comment
+    my $pkg = __PACKAGE__;
+    my $now = localtime();
+
+    my $header = $doc->createComment( <<_HEADER . '    ' );
+ BE WARNED: in most cases, the example below cannot be used without
+  interpretation. The comments will guide you.
+  Produced by $pkg version $VERSION on $now
+_HEADER
+
+    unless($args{skip_header})
+    {   $xml->insertBefore($header, $xml->firstChild);
+        $xml->insertBefore($doc->createTextNode("\n  "), $header);
+    }
+
+    # I use xsi:type myself, too late for the usual "used" counter
+    $ast->{'xmlns:xsi'} ||= SCHEMA2001i
+        if $args{show_type};
+
+    # add info about name-spaces
+    foreach (sort keys %$ast)
+    {   if( m/^xmlns\:(.*)/ )
+        {   $xml->setNamespace($ast->{$_}, $1, 0);
+        }
+    }
+
+    $xml;
 }
 
 sub _xmlAny($$$$);
 sub _xmlAny($$$$)
 {   my ($self, $doc, $ast, $indent, $args) = @_;
     my @res;
+    my $xsi = $self->_registerNSprefix('xsi', SCHEMA2001i, 1)
+        if $args->{show_type};
 
     my @comment;
     if($ast->{struct} && $args->{show_struct})
@@ -596,8 +870,13 @@ sub _xmlAny($$$$)
         push @comment, ref $struct ? @$struct : $struct;
     }
 
-    push @comment, $ast->{occur}  if $ast->{occur}  && $args->{show_occur};
-    push @comment, $ast->{facets} if $ast->{facets} && $args->{show_facets};
+    push @comment, $ast->{occur}
+        if $ast->{occur}  && $args->{show_occur};
+
+    if($ast->{facets}  && $args->{show_facets})
+    {   my $facets = $ast->{facets};
+        push @comment, ref $facets eq 'ARRAY' ? @$facets : $facets;
+    }
 
     if(defined $ast->{kind} && $ast->{kind} eq 'union')
     {   push @comment, map { "  $_->{type}"} @{$ast->{choice}};
@@ -639,11 +918,9 @@ sub _xmlAny($$$$)
           (@comment ? "$indent$example$outdent" : $example)
     }
 
-    if($ast->{type} && $args->{show_type})
-    {   my $full = $ast->{type};
-        my ($ns, $type) = unpack_type $full;
-        # Don't known how to encode the namespace (yet)
-        push @res, $doc->createAttribute(type => $type);
+    if($ast->{_TYPE} && $args->{show_type})
+    {   my $pref = $self->prefixed($ast->{_TYPE});
+        push @res, $doc->createAttribute("$xsi:type" => $pref);
     }
 
     return @res
@@ -655,6 +932,11 @@ sub _xmlAny($$$$)
     $node;
 }
 
+sub makeBlocked($$$)
+{   my ($self, $where, $class, $type) = @_;
+    panic "namespace blocking not yet supported for Templates";
+}
+
 =chapter DETAILS
 
 =section Processing Wildcards
@@ -665,6 +947,16 @@ Hooks are implemented since version 0.82.  They can be used to
 improve the template output.
 
 =subsection hooks executed before the template is generated
+None defined yet.
+
+=subsection hooks executed as replacement
+
+The predefined hook C<COLLAPSE> can be used to remove the extensive
+listing of some elements.  Usually used with a type of which you
+know the structure or which is repeated often.
+
+=subsection hooks for post-processing, after the data is collected
+None defined yet.
 
 =section Typemaps
 Typemaps are currently only available to improve the PERL output.
